@@ -15,9 +15,9 @@ try:
 except ImportError:
     pass
 
-from . import digitalocean as do
+from . import digitalocean, dynadot
 
-__all__ = ["main", "Gaea"]
+__all__ = ["digitalocean", "dynadot", "Gaea", "main"]
 
 LOGO = """\
  ██████╗  █████╗ ███████╗ █████╗ 
@@ -85,13 +85,52 @@ VERSIONS = {
 }
 
 
+def build_linux():
+    """
+    Compile a standalone binary for distribution on linux systems.
+
+    Upload binary to the bootstrap server upon successful compilation.
+
+    Call from inside the project's working directory using `poetry run build_linux`
+
+    """
+    if sh.poetry(
+        "run",
+        "python",
+        "-m",
+        "nuitka",
+        "main.py",
+        "--plugin-enable=pyside6",
+        "--plugin-enable=gevent",
+        "--onefile",
+        "-o",
+        "dist/gaea",
+        "--output-dir=dist",
+        "--linux-onefile-icon=icon.png",
+    ).exit_code:
+        print("build failed!")
+    else:
+        sh.scp(
+            "-i", "gaea_key", "./dist/gaea", "root@137.184.4.213:/var/www/html/dists/"
+        )
+        print("build uploaded!")
+
+
+def build_windows():
+    pass
+
+
+def build_macos():
+    pass
+
+
 def spawn_cloud(name, digitalocean_token):
     """Spawn the understory on a new droplet."""
-    cli = do.Client(digitalocean_token)
-    key = do.get_key(cli)
+    cli = digitalocean.Client(digitalocean_token)
+    key = digitalocean.get_key(cli)
     print("generating droplet:", end=" ", flush=True)
     droplet = cli.create_droplet(name, size="1gb", ssh_keys=[key["id"]])
-    do.wait(cli, droplet["id"])
+    digitalocean.wait(cli, droplet["id"])
     droplet = cli.get_droplet(droplet["id"])
     ip_address = None
     for ip_details in droplet["networks"]["v4"]:
@@ -101,7 +140,7 @@ def spawn_cloud(name, digitalocean_token):
     print(ip_address, flush=True)
 
     print("spawning gaea..", flush=True)
-    root_ssh = do.get_ssh("root", ip_address)
+    root_ssh = digitalocean.get_ssh("root", ip_address)
     root_ssh("adduser gaea --disabled-login --gecos gaea")
     root_ssh('echo "gaea  ALL=NOPASSWD: ALL" | tee -a /etc/sudoers.d/01_gaea')
 
@@ -114,27 +153,28 @@ def spawn_cloud(name, digitalocean_token):
 
     print("spawning the understory..", flush=True)
     print("__file__:", __file__)
-    do.scp("gaea", f"gaea@{ip_address}:")
-    gaea_ssh = do.get_ssh("gaea", ip_address)
+    digitalocean.scp("gaea", f"gaea@{ip_address}:")
+    gaea_ssh = digitalocean.get_ssh("gaea", ip_address)
     gaea_ssh("chown gaea:gaea gaea")
     gaea_ssh("chmod u+x gaea")
     gaea_ssh(f"./gaea setup {digitalocean_token}", process_out(ip_address))
 
 
-def spawn_local():
-    # sh.sudo("bash", "spawn.sh")  # TODO FIXME
-
-    # sh.scp("-i", "gaea_key", "gaea", "gaea@localhost:")
-
-    # TODO sh bake
-    def ssh(command):
-        sh.ssh(
-            "gaea@localhost", "-i", "gaea_key", command, _out=process_out("localhost")
-        )
-
-    # ssh("chown gaea:gaea gaea")
-    # ssh("chmod u+x gaea")
-    ssh("./gaea setup")
+# TODO XXX def spawn_local():
+# TODO XXX     # sh.sudo("bash", "spawn.sh")  # TODO FIXME
+# TODO XXX
+# TODO XXX     # sh.scp("-i", "gaea_key", "gaea", "gaea@localhost:")
+# TODO XXX
+# TODO XXX     # TODO sh bake
+# TODO XXX     def ssh(command):
+# TODO XXX         sh.ssh(
+# TODO XXX             "gaea@localhost", "-i", "gaea_key", command,
+# TODO XXX             _out=process_out("localhost")
+# TODO XXX         )
+# TODO XXX
+# TODO XXX     # ssh("chown gaea:gaea gaea")
+# TODO XXX     # ssh("chmod u+x gaea")
+# TODO XXX     ssh("./gaea setup")
 
 
 def process_out(host):
@@ -197,6 +237,74 @@ def log(*args, **kwargs):
         print(*args, **kwargs)
 
 
+def get_ip():
+    return sh.hostname("-I").split()[0]
+
+
+def install_app(app, home_dir, github_token):
+    if app.startswith("https://"):  # Git repo
+        pass  # TODO
+    elif "/" in app:  # GitHub project
+        working_dir = pathlib.Path(home_dir) / "working"
+        user, _, repo = app.partition("/")
+        sh.mkdir(working_dir / user, "-p")
+        sh.git("clone", f"https://github.com/{app}.git", _cwd=working_dir / user)
+        sh.Command(f"{home_dir}/.local/bin/poetry")(
+            "install",
+            _cwd=working_dir / user / repo,
+            _env={},  # NOTE ensures VIRTUAL_ENV is absent
+        )
+        # XXX gh_prefix = f"https://api.github.com/repos/{app}/releases"
+        # XXX gh_token_header = f"token {github_token}"
+        # XXX gh_headers = {
+        # XXX     "Accept": "application/vnd.github.v3+json",
+        # XXX     "Authorization": gh_token_header,
+        # XXX }
+        # XXX asset = web.get(f"{gh_prefix}/latest",
+        # XXX                 headers=gh_headers).json["assets"][0]
+        # XXX sh.wget(
+        # XXX     f"{gh_prefix}/assets/{asset['id']}",
+        # XXX     "-O",
+        # XXX     asset["name"],
+        # XXX     "--header",
+        # XXX     "Accept: application/octet-stream",
+        # XXX     "--header",
+        # XXX     f"Authorization: {gh_token_header}",
+        # XXX     _cwd=home_dir,
+        # XXX )
+        # XXX sh.sh(
+        # XXX     "runinenv",
+        # XXX     "system/env",
+        # XXX     "pip",
+        # XXX     "install",
+        # XXX     # TODO "--index-url",
+        # XXX     # TODO "",
+        # XXX     # TODO "--extra-index-url",
+        # XXX     # TODO "",
+        # XXX     asset["name"],
+        # XXX     _cwd=home_dir,
+        # XXX )
+    else:  # PyPI package
+        sh.sh(
+            "runinenv",
+            "system/env",
+            "pip",
+            "install",
+            # TODO "--index-url",
+            # TODO "",
+            # TODO "--extra-index-url",
+            # TODO "",
+            app,
+            _cwd=home_dir,
+        )
+    sh.sudo("supervisorctl", "restart", "gaea-app", "gaea-app-jobs")
+
+
+def upgrade_app(app, home_dir):
+    sh.sh("runinenv", "system/env", "pip", "install", "-U", app, _cwd=home_dir)
+    sh.sudo("supervisorctl", "restart", "gaea-app")
+
+
 class Gaea:
     """The host of the machine."""
 
@@ -267,7 +375,7 @@ class Gaea:
 
         if self.env_dir.exists():
             return
-        log("creating virtual environment")
+        log("creating primary virtual environment")
         get_python_sh()("-m", "venv", self.env_dir)
         sh.echo(
             textwrap.dedent(
@@ -282,37 +390,38 @@ class Gaea:
             _out=f"{self.home_dir}/runinenv",
         )
         sh.chmod("+x", self.home_dir / "runinenv")
-        sh.sh(self.home_dir / "runinenv", self.env_dir, "pip", "install", "sh==1.11")
 
-        log("installing SQLite")
-        log(" ", "downloading")
-        sh.wget("https://www.sqlite.org/src/tarball/sqlite.tar.gz", _cwd=self.src_dir)
-        log(" ", "extracting")
-        sh.tar("xf", "sqlite.tar.gz", _cwd=self.src_dir)
-        sqlite_dir = self.src_dir / "sqlite"
-        log(" ", "configuring")
-        sh.bash("./configure", _cwd=sqlite_dir)
-        sh.make("sqlite3.c", _cwd=sqlite_dir)
-        sh.git("clone", "https://github.com/coleifer/pysqlite3", _cwd=self.src_dir)
-        pysqlite_dir = self.src_dir / "pysqlite3"
-        sh.cp(sqlite_dir / "sqlite3.c", ".", _cwd=pysqlite_dir)
-        sh.cp(sqlite_dir / "sqlite3.h", ".", _cwd=pysqlite_dir)
-        sh.sh(
-            self.home_dir / "runinenv",
-            self.env_dir,
-            "python",
-            "setup.py",
-            "build_static",
-            _cwd=pysqlite_dir,
-        )
-        sh.sh(
-            self.home_dir / "runinenv",
-            self.env_dir,
-            "python",
-            "setup.py",
-            "install",
-            _cwd=pysqlite_dir,
-        )
+        # sh.sh(self.home_dir / "runinenv", self.env_dir, "pip", "install", "sh==1.11")
+
+        # log("installing SQLite")
+        # log(" ", "downloading")
+        # sh.wget("https://www.sqlite.org/src/tarball/sqlite.tar.gz", _cwd=self.src_dir)
+        # log(" ", "extracting")
+        # sh.tar("xf", "sqlite.tar.gz", _cwd=self.src_dir)
+        # sqlite_dir = self.src_dir / "sqlite"
+        # log(" ", "configuring")
+        # sh.bash("./configure", _cwd=sqlite_dir)
+        # sh.make("sqlite3.c", _cwd=sqlite_dir)
+        # sh.git("clone", "https://github.com/coleifer/pysqlite3", _cwd=self.src_dir)
+        # pysqlite_dir = self.src_dir / "pysqlite3"
+        # sh.cp(sqlite_dir / "sqlite3.c", ".", _cwd=pysqlite_dir)
+        # sh.cp(sqlite_dir / "sqlite3.h", ".", _cwd=pysqlite_dir)
+        # sh.sh(
+        #     self.home_dir / "runinenv",
+        #     self.env_dir,
+        #     "python",
+        #     "setup.py",
+        #     "build_static",
+        #     _cwd=pysqlite_dir,
+        # )
+        # sh.sh(
+        #     self.home_dir / "runinenv",
+        #     self.env_dir,
+        #     "python",
+        #     "setup.py",
+        #     "install",
+        #     _cwd=pysqlite_dir,
+        # )
 
         log("installing Gaea")
         sh.sh(
@@ -323,6 +432,18 @@ class Gaea:
             "libgaea",
         )
 
+        log("installing Poetry")
+        get_python_sh()(
+            sh.wget(
+                "https://raw.githubusercontent.com/python-poetry/poetry"
+                "/master/install-poetry.py",
+                "-q",
+                "-O",
+                "-",
+            ),
+            "-",
+        )
+
     def setup_nginx(self):
         """Install Nginx (w/ TLS, HTTPv2, RTMP) for web serving."""
         nginx_src = f"nginx-{VERSIONS['nginx']}"
@@ -492,7 +613,11 @@ class Gaea:
     def mount_site(self, site, app):
         """Instruct Nginx to route requests for `site` to `app`."""
         if app not in get_statuses():
-            self.run_app(app)
+            # XXX if site.startswith("dev."):
+            if "#" in app:
+                self.run_development_app(app)
+            else:
+                self.run_production_app(app)
         config = self.get_config()
         config["websites"][site] = app
         self.save_config(config)
@@ -504,7 +629,44 @@ class Gaea:
             )
         sh.sudo("supervisorctl", "restart", "nginx")
 
-    def run_app(self, app, workers=2):
+    def run_development_app(self, app, workers=2):
+        """Instruct Supervisor to run given `app` using gunicorn."""
+        supervisor = configparser.ConfigParser()
+        project, _, callable = app.partition("#")
+        name = callable.replace(":", "-")
+        app_dir = self.apps_dir / project / name
+        app_dir.mkdir(parents=True, exist_ok=True)
+        env_dir = sh.Command(f"{self.home_dir}/.local/bin/poetry")(
+            "env",
+            "info",
+            "--path",
+            _cwd=self.working_dir / project,
+            _env={},  # NOTE ensures VIRTUAL_ENV is absent
+        )
+        dashed_project = project.replace("/", "-")
+        command = (
+            f"{self.home_dir}/runinenv {env_dir} gunicorn {callable}"
+            f" -k gevent -w {workers} --bind unix:{app_dir}/app.sock"
+        )
+        supervisor[f"program:{dashed_project}-{name}"] = {
+            "autostart": "true",
+            "command": command,
+            "directory": app_dir,
+            "environment": "PYTHONUNBUFFERED=1",
+            "stopsignal": "INT",
+            "user": "gaea",
+        }
+        command = f"{self.home_dir}/runinenv {env_dir} loveliness serve"
+        supervisor[f"program:{dashed_project}-{name}-jobs"] = {
+            "autostart": "true",
+            "command": command,
+            "directory": app_dir,
+            "stopsignal": "INT",
+            "user": "gaea",
+        }
+        self._write_supervisor_conf(f"app-{dashed_project}-{name}", supervisor)
+
+    def run_production_app(self, app, workers=2):
         """Instruct Supervisor to run given `app` using gunicorn."""
         supervisor = configparser.ConfigParser()
         name = app.replace(":", "-")
